
import javax.vecmath.*;
import javax.media.j3d.*;
import com.sun.j3d.utils.geometry.*;
/**
 * ElevationModel is a specialization of a BranchGroup that holds Java3D
 * geometry objects representing a terrain map. The map is divided  into segments
 * each segment contains a Level-of-Detail switch  containing the geometry
 * for the segment in three levels of detail. This allows for faster drawing and
 * navigation of the model as a whole.
 *
 * @author  Mark Pendergast
 * @version 1.0 February 2003
 * @see LODSegment
*/

public class ElevationModel extends BranchGroup implements ElevationModelInterface
{
/** ground coordinates for the model, contains the lat/long of the four corners in arc seconds */
public GroundCoordinates groundCoordinates;

 /** minimum X coordinate */
 private float west_X = 0;
 /** maximum X coordinate */
 private float east_X = 0;
 /** maximum Z  coordinate, note, positive Z axis is in the south due to Java3D */
 private  float south_Z = 0;
 /** minimum Z coordinate */
private  float north_Z = 0;
/** number of seconds high/wide of each segment is */
 private static int SECONDS_PER_SEGMENT = 600;
/** space in meters between each row of elevation data */
 private float deltaRow;
/** space in meters between each column of elevation data */
 private float deltaCol;
 /** factor each elevation reading is multiplied by when drawing the model. Makes elevation differences stand out */
 public float exageration = 3;
 /** minimum elevation in meters */
 public int minElevation ;
 /** maximum elevation in meters */
 public int maxElevation;
/** geographic name or title of the data */
 public  String name;
 /** two dimensional array of segments comprising the model data */
  private LODSegment[][] segments;
  /**
   * Load the DEM file and create a set of LODSegments based on its data.
   *
   * @param aFileName name of the DEM file
   * @param stat reference to a StatusWindow object used to display load progress
   *
   */
 public ElevationModel(String aFileName,StatusWindow stat) throws IllegalArgumentException
 {
 int sRows, sColumns;
 float length, width;

  setCapability(BranchGroup.ALLOW_DETACH);  // allow this model to be removed from the Universe
 //
 // for now, just use one file and its ground coordinates
 //
  if(stat != null)
     stat.setLabel2("Loading DEM data ");
  DemFile  file =  new DemFile(aFileName);

   groundCoordinates = file.groundCoordinates;
   minElevation = file.minElevation;
   maxElevation = file.maxElevation;
   name = file.quadrangleName;
   length = groundCoordinates.lengthMeters();
   width = groundCoordinates.widthMeters();
   west_X = (-length/2) ;
   east_X = (length/2);
   south_Z = (width/2);
   north_Z = (-width/2);
  //
 //   create LODSegments
 //
   sColumns = (int)Math.ceil(groundCoordinates.lengthSeconds()/SECONDS_PER_SEGMENT);
   sRows = (int)Math.ceil(groundCoordinates.widthSeconds()/SECONDS_PER_SEGMENT);
   segments = new LODSegment[sRows][sColumns];

   GroundCoordinates gc = new GroundCoordinates();

   int rowRatio = (int) (1.0d*file.nRows/sRows);
   int colRatio = (int) (1.0d*file.nColumns/sColumns);

   deltaRow = (north_Z-south_Z)/sRows;
   deltaCol = (east_X-west_X)/sColumns;
   for(int row = 0; row < sRows; row++)
   {
     float minX, maxX, minZ, maxZ;
     int startRow, stopRow, startCol, stopCol;
     gc.sw[GroundCoordinates.LATITUDE] = groundCoordinates.sw[GroundCoordinates.LATITUDE] + row*SECONDS_PER_SEGMENT;
     gc.se[GroundCoordinates.LATITUDE] = groundCoordinates.sw[GroundCoordinates.LATITUDE]+ row*SECONDS_PER_SEGMENT;
     gc.nw[GroundCoordinates.LATITUDE] = groundCoordinates.sw[GroundCoordinates.LATITUDE] + (row+1)*SECONDS_PER_SEGMENT;
     gc.ne[GroundCoordinates.LATITUDE] = groundCoordinates.sw[GroundCoordinates.LATITUDE]+ (row+1)*SECONDS_PER_SEGMENT;
     minZ = south_Z + row*deltaRow;
     maxZ = south_Z +(row+1.0f)*deltaRow;
     startRow = row*(rowRatio);
     stopRow = (row+1)*(rowRatio);
     for(int col = 0 ; col < sColumns; col++)
     {
      if(stat != null)
       stat.setLabel2("Creating geometry segment ",row*sColumns+col+1,sRows*sColumns);
     minX = west_X + col*deltaCol;
     maxX = west_X +(col+1.0f)*deltaCol;
     startCol = col*(colRatio);
     stopCol = (col+1)*(colRatio);
     gc.sw[GroundCoordinates.LONGITUDE] = groundCoordinates.sw[GroundCoordinates.LONGITUDE] - col*SECONDS_PER_SEGMENT;
     gc.nw[GroundCoordinates.LONGITUDE] = groundCoordinates.sw[GroundCoordinates.LONGITUDE]- col*SECONDS_PER_SEGMENT;
     gc.se[GroundCoordinates.LONGITUDE] = groundCoordinates.sw[GroundCoordinates.LONGITUDE] - (col+1)*SECONDS_PER_SEGMENT;
     gc.ne[GroundCoordinates.LONGITUDE] = groundCoordinates.sw[GroundCoordinates.LONGITUDE]- (col+1)*SECONDS_PER_SEGMENT;

     segments[row][col] = new LODSegment(file.elevations,  startRow,startCol, stopRow,stopCol,
                                                 minElevation, maxElevation, gc, exageration,
                                                 minX, maxX, minZ, maxZ);
      addChild(segments[row][col]);
    }
 }
//
//  Adjust normals on borders of each segment to remove variations
//  Stitch right to left first
//
  if(stat != null)
     stat.setLabel2("Adjusting normals in adjacent segments");
   for(int row = 0; row < sRows; row++)
   {
    for(int col = 0 ; col < sColumns; col++)
     {
       if(col < sColumns-1) // only if not last column
	     segments[row][col].stitchRight(segments[row][col+1]);
    }
  }
  //
  //Now stitch tops to bottoms, note this must be done in a separate
  // loop so that the corners come out ok.
  //
     for(int row = 0; row < sRows; row++)
     {
      for(int col = 0 ; col < sColumns; col++)
       {
         if(row < sRows-1) // only if not last row
  	     segments[row][col].stitchTop(segments[row+1][col]);
      }
  }
  //
  if(stat != null)
     stat.setLabel2("Compiling/Optimizing the geometry");

  compile(); // compile the model
 }
/**
* Fetches the elevation at a particular location on the terrain map given
* the x,z coordinates.  X,Z coordinates represent the distance in meters
* from the center of the terrain image
* @param x x coordinate
* @param z z coordinate
* @return the elevation (y coordinate) in meters (adjusted for exageration)
*/
 public float getElevationAt(float x, float z)
 {
   int row = (int)Math.abs((z-south_Z)/deltaRow);
   int col = (int)Math.abs((x-west_X)/deltaCol);

   row = Math.min(row,segments.length-1);
   col = Math.min(col, segments[0].length-1);
   row = Math.max(row,0);
   col = Math.max(col, 0);
   return segments[row][col].getElevationAt(x,z);
  }
/**
 * retrieve the model length, distance in meters from west to east
 *  @return the length in meters
 */
  public float getModelLength()
  {
	return groundCoordinates.lengthMeters();
 }
/**
 * retrieve the model width, distance in meters from south to north
 *  @return the width in meters
 */
  public float getModelWidth()
  {
	return groundCoordinates.widthMeters();
 }
/**
 * retrieve the model maximum elevation, adjusted by the elevation exageration
 *  @return the adjusted maximum elevation in meters
 */
  public float getMaxElevation()
  {
	return maxElevation*exageration;
 }
/**
 * retrieve the model minimum elevation, adjusted by the elevation exageration
 *  @return the adjusted minimum elevation in meters
 */
  public float getMinElevation()
  {
	return minElevation*exageration;
 }
}