import javax.vecmath.*;
import javax.media.j3d.*;
import com.sun.j3d.utils.geometry.*;
/**
 * ElevationSegment is a specialization of Shape3d. It creates a
 * 3D map of terrain data using an interleaved triangle strip array.
 *  Terrain data is passed into ElevationSegment as a two dimensional
 *  array of elevations.  These are converted into a series of triangle
 * strips to represent their geometry.
 *
 * @author  Mark Pendergast
 * @version 1.0 February 2003
 */

public class ElevationSegment extends Shape3D {

/** ground coordinates for the model, contains the lat/long of the four corners in arc seconds */
public GroundCoordinates groundCoordinates;
/** number of rows of vertex data stored */
private int dRows;
/** number of columns of vertex data stored */
private int dColumns;
/** meters between columns of data */
private float deltaX=0;
/** meters between rows of data */
private float deltaZ=0;
/** first column logical coordinate*/
private float xStart=0;
/** first row logical coordinate*/
private float zStart=0;

//
// vertex data holds 9 doubles per vertex
// color, normal, and coordinate
//
private static  final int FLOATSPERVERTEX= 9;
private static final int COLOR_OFFSET = 0;
private static final int NORMAL_OFFSET = 3;
private  static final int COORD_OFFSET = 6;
private float[] vertexData;

private static Color3f matColor = new Color3f(.9f,.7f,.2f);
private InterleavedTriangleStripArray  tStrip;
/**
* Constructor is reponsible for setting up the appearance/material,
* computing the triangle strip values (colors, normals, coordinates)
* stored in vertexData, then creating the actual JAVA 3D geometry.
*
*  @param elevations two dimensional array of terrain elevation data in meters
*  @param startRow first row of data to use from elevations
*  @param startColumn first column of data to use from elevations
*  @param stopRow last row of data to use from elevations
*  @param stopColumn last column of data to use from elevations
*  @param minEl minimum elevation in elevations
*  @param maxEl maximum elevation in elevations
*  @param gc latitude, longitude of the corners of this segment
*  @param exageration amount to exagerate(multiply) elevations by
*  @param lowX starting x coordinate in meters
*  @param highX stopping x coordinate in meters
*  @param lowZ starting z coordinate in meters
*  @param highZ stopping z coordinate in meters
*  @param resolution number of rows/columns of elevation data to skip when mapping. A resolution
*  of 1 implies mapping every elevation point, a resolution of 10 implies mapping every 10th point.
*/
public ElevationSegment( int elevations[][],   int startRow, int startColumn, int stopRow, int stopColumn, int minEl, int maxEl, GroundCoordinates gc,
             float exageration, float lowX, float highX, float lowZ, float highZ, int resolution)
{
//
// save the ground coordinates
//
  groundCoordinates = gc;
 //
 // setup material properties
 //
 setupAppearance();
//
// process the 2D elevation array
//
  dRows = (int)Math.ceil((stopRow-startRow+1)/(double)resolution);
  dColumns = (int)Math.ceil((stopColumn-startColumn+1)/(double)resolution);
  xStart = lowX;
  zStart = lowZ;
  deltaX = (highX-lowX)/(stopColumn-startColumn);
  deltaZ = (highZ-lowZ)/(stopRow-startRow);
//
// first create an interleaved array of colors,  normals, and points
//
  vertexData = new float[FLOATSPERVERTEX*(dRows)*2*(dColumns-1)];
  if(vertexData == null)
  {
    System.out.println("Elevation segment: memory allocation failure");
    return;
  }
//
// populate vertexData a strip at a time
//
  int row, col; // used as indexes into the elevations array
  int i; // used as an index into vertexData

  for( col = startColumn, i = 0; col <= stopColumn-resolution; col += resolution)
  {
   for(row = startRow; row <= stopRow; row += resolution)
   {
	 if(row+resolution > stopRow) // always use last data line to prevent seams
	    row = stopRow;

    setColor(i+COLOR_OFFSET,elevations[row][col],minEl,maxEl);
   	setCoordinate(elevations, i+COORD_OFFSET,row,col,startRow,startColumn,exageration);
	i += FLOATSPERVERTEX;

	int c = col;
	if(c+resolution > stopColumn-resolution) // always use last data line to prevent seams
     c = stopColumn-resolution;

    setColor(i+COLOR_OFFSET,elevations[row][c+resolution],minEl,maxEl);
	setCoordinate(elevations, i+COORD_OFFSET,row,c+resolution,startRow,startColumn,exageration);
	i += FLOATSPERVERTEX;
   }
  }

//
// create a stripCount array showing the number of vertices in each
// strip.
//
 int[] stripCounts = new int[dColumns-1];
 for(int strip = 0; strip < dColumns-1; strip++)
   stripCounts[strip] = (dRows)*2;

//
// Create and set the geometry
//

tStrip = new InterleavedTriangleStripArray(vertexData.length/FLOATSPERVERTEX,
      GeometryArray.COORDINATES|GeometryArray.COLOR_3|GeometryArray.NORMALS|GeometryArray.BY_REFERENCE|GeometryArray.INTERLEAVED,
      stripCounts);
    tStrip.setInterleavedVertices(vertexData);
    tStrip.generateNormals(true);
    setGeometry(tStrip);

}
/**
 *  setup the material properties and coloring attributes
 *
 */
private void setupAppearance()
{
    Appearance app = new Appearance(); // create an appeance
    Material mat = new Material(); // create a material
    ColoringAttributes ca = new ColoringAttributes(matColor,ColoringAttributes.SHADE_GOURAUD); // selecting shading
    app.setColoringAttributes(ca); // add coloring attributes to the appearance

    mat.setLightingEnable(true); // allow lighting
    mat.setDiffuseColor(matColor); // set diffuse color (used by directional lights)
    mat.setAmbientColor(matColor); // set ambient color (used by ambient lights)
    mat.setSpecularColor(0f,0f,0f); // no specular color
    mat.setShininess(1.0f); // no shininess
    mat.setLightingEnable(true); // allow lighting

    app.setMaterial(mat); // add the material to the appearance
    setAppearance(app); // add the appearance to the object
    setCapability(Shape3D.ALLOW_GEOMETRY_WRITE|Shape3D.ALLOW_GEOMETRY_READ); // allows calls to setgeometry
}
/**
 * store coordinate data into vertex data array
 *
 *  @param elevations array of elevations
 *  @param i index into vertexData to store coordinate
 *  @param row elevation row
 * @param col elevation column
 * @param startRow first row used in elevations
 *  @param stopColumn first column used in elevations
 *  @param exageration elevation exageration factor
 */
 public void setCoordinate(int[][] elevations, int i, int row, int col, int startRow, int startColumn,float exageration)
 {
    vertexData[i] = (float)(((col-startColumn)*deltaX)+xStart);
   vertexData[i+1] = elevations[row][col]*exageration;
   vertexData[i+2] = (float)(zStart+((row-startRow)*deltaZ));
 }
/**
 * store color data into vertex data array, compute color based
 * on the elevation's distance between min and max elevations
 *
 *  @param i index into vertexData to store coordinate
 *  @param elevation  vertex elevation (no exageration)
 *  @param minElevation minimum elevation in model
 *  @param maxElevation maximum elevation in model
 */

public void setColor(int i, int elevation,int minElevation, int maxElevation)
{
  float ratio = ((float)	elevation)/((float)maxElevation-minElevation);

vertexData[i] = matColor.x*ratio; // set red
vertexData[i+1] = matColor.y*ratio; // set green
vertexData[i+2] = (float)(1-ratio); // trick to bring blue for the lowest elevations
}
/**
*  stitch right of this segment to left of other segment, stitching
*  consists of averaging the normals on the common boundary vertices
* between segments. That is, since adjacent segments have the same
* vertices in common (the left side of one is the same as the right side of
* the other), then these normals should be the same as well to prevent
*  visible seams.
*
*  @param other reference to the segment to the right of this one
*/
public void stitchRight (ElevationSegment other)
{
  //
  // average the normals of the first left side column of the other segment
  // to the right side of the last column of this segment
  //
      int j = 0;    // j starts  on left side of first column
     int i = FLOATSPERVERTEX*2*dRows*(dColumns-2)+FLOATSPERVERTEX; // right side of last column
     for(int row = 0; row < dRows; row++)
     {
      if(vertexData[i+COORD_OFFSET] == other.vertexData[j+COORD_OFFSET] &&
         vertexData[i+COORD_OFFSET+1] == other.vertexData[j+COORD_OFFSET+1] &&
         vertexData[i+COORD_OFFSET+2] == other.vertexData[j+COORD_OFFSET+2])
      {
	    vertexData[i+NORMAL_OFFSET] = (vertexData[i+NORMAL_OFFSET]+other.vertexData[j+NORMAL_OFFSET])/2.0f;
	    vertexData[i+NORMAL_OFFSET+1] = (vertexData[i+NORMAL_OFFSET+1]+other.vertexData[j+NORMAL_OFFSET+1])/2.0f;
	    vertexData[i+NORMAL_OFFSET+2] = (vertexData[i+NORMAL_OFFSET+2]+other.vertexData[j+NORMAL_OFFSET+2])/2.0f;
         tStrip.normalize(vertexData,i);
	    other.vertexData[j+NORMAL_OFFSET] = vertexData[i+NORMAL_OFFSET];
	    other.vertexData[j+NORMAL_OFFSET+1] = vertexData[i+NORMAL_OFFSET+1];
	    other.vertexData[j+NORMAL_OFFSET+2] = vertexData[i+NORMAL_OFFSET+2];
     }
      i+= FLOATSPERVERTEX*2; // increment to next row
      j+= FLOATSPERVERTEX*2; // increment to next row
     } // end for rows
}
/**
*  stitch top of this segment to bottom of other segment, stitching
*  consists of averaging the normals on the common boundary vertices
* between segments. That is, since adjacent segments have the same
* vertices in common (the top of one is the same as the bottom of
* the other), then these normals should be the same as well to prevent
*  visible seams.
*
*  @param other reference to the segment to the right of this one
*/
public void stitchTop(ElevationSegment other)
{
  //
 //
     int i = FLOATSPERVERTEX*2*(dRows-1);    // i starts  on right side top of first strip
     int j = 0; // right side bottom of first strip
     for(int strip = 0; strip < dColumns-1; strip++)
     {
	  // do right side of strip
      if(vertexData[i+COORD_OFFSET] == other.vertexData[j+COORD_OFFSET] &&
         vertexData[i+COORD_OFFSET+1] == other.vertexData[j+COORD_OFFSET+1] &&
         vertexData[i+COORD_OFFSET+2] == other.vertexData[j+COORD_OFFSET+2])
      {
	    vertexData[i+NORMAL_OFFSET] = (vertexData[i+NORMAL_OFFSET]+other.vertexData[j+NORMAL_OFFSET])/2.0f;
	    vertexData[i+NORMAL_OFFSET+1] = (vertexData[i+NORMAL_OFFSET+1]+other.vertexData[j+NORMAL_OFFSET+1])/2.0f;
	    vertexData[i+NORMAL_OFFSET+2] = (vertexData[i+NORMAL_OFFSET+2]+other.vertexData[j+NORMAL_OFFSET+2])/2.0f;
        tStrip.normalize(vertexData,i);
        other.vertexData[j+NORMAL_OFFSET] = vertexData[i+NORMAL_OFFSET];
	    other.vertexData[j+NORMAL_OFFSET+1] = vertexData[i+NORMAL_OFFSET+1];
	    other.vertexData[j+NORMAL_OFFSET+2] = vertexData[i+NORMAL_OFFSET+2];
   }

    // do left side of strip
    if(vertexData[i+COORD_OFFSET+FLOATSPERVERTEX] == other.vertexData[j+COORD_OFFSET+FLOATSPERVERTEX] &&
         vertexData[i+COORD_OFFSET+1+FLOATSPERVERTEX] == other.vertexData[j+COORD_OFFSET+1+FLOATSPERVERTEX] &&
         vertexData[i+COORD_OFFSET+2+FLOATSPERVERTEX] == other.vertexData[j+COORD_OFFSET+2+FLOATSPERVERTEX])
      {
	    vertexData[i+NORMAL_OFFSET+FLOATSPERVERTEX] = (vertexData[i+NORMAL_OFFSET+FLOATSPERVERTEX]+other.vertexData[j+NORMAL_OFFSET+FLOATSPERVERTEX])/2.0f;
	    vertexData[i+NORMAL_OFFSET+1+FLOATSPERVERTEX] = (vertexData[i+NORMAL_OFFSET+1+FLOATSPERVERTEX]+other.vertexData[j+NORMAL_OFFSET+1+FLOATSPERVERTEX])/2.0f;
	    vertexData[i+NORMAL_OFFSET+2+FLOATSPERVERTEX] = (vertexData[i+NORMAL_OFFSET+2+FLOATSPERVERTEX]+other.vertexData[j+NORMAL_OFFSET+2+FLOATSPERVERTEX])/2.0f;
        tStrip.normalize(vertexData,i+FLOATSPERVERTEX);
	    other.vertexData[j+NORMAL_OFFSET+FLOATSPERVERTEX] = vertexData[i+NORMAL_OFFSET+FLOATSPERVERTEX];
	    other.vertexData[j+NORMAL_OFFSET+1+FLOATSPERVERTEX] = vertexData[i+NORMAL_OFFSET+1+FLOATSPERVERTEX];
	    other.vertexData[j+NORMAL_OFFSET+2+FLOATSPERVERTEX] = vertexData[i+NORMAL_OFFSET+2+FLOATSPERVERTEX];
    }

      i+= FLOATSPERVERTEX*2*dRows; // increment to next strip
      j+= FLOATSPERVERTEX*2*dRows; // increment to next strip
     } // end for rows
}
/**
* Fetches the elevation at a particular location on the terrain map given
* the x,z coordinates.  X,Z coordinates represent the distance in meters
* from the center of the terrain image
* @param x x coordinate
* @param z z coordinate
* @return the elevation (y coordinate) in meters (adjusted for exageration)
*/
 public float getElevationAt(float x, float z)
 {
	 int col= (int)Math.abs((x-xStart)/deltaX);
	 int row = (int)Math.abs((z-zStart)/deltaZ);

     row = Math.min(row,dRows-1);
    col = Math.min(col, dColumns-2);
    row = Math.max(row,0);
    col = Math.max(col, 0);

	 int i = (FLOATSPERVERTEX*2*dRows)*col + (FLOATSPERVERTEX*2)*row;
	 return vertexData[i+COORD_OFFSET+1];
 }


}